"""
Module for a Response object.

A Response is returned by Routes when the underlying coroutine is done.
"""
import gzip
import http
import typing
import warnings

try:
    from http_parser.parser import IOrderedDict
except ImportError:  # pragma: no cover
    from http_parser.pyparser import IOrderedDict

    warnings.warn("Using fallback Python HTTP parser - this will negatively affect performance.")

from . import util

try:
    import magic
except (ImportError, OSError):  # pragma: no cover
    _has_magic = False
    warnings.warn("Cannot load libmagic - Cannot determine file types automatically...")
else:
    _has_magic = True

from email.utils import formatdate
from http.cookies import SimpleCookie


class Response(object):
    """
    A response is responsible (no pun intended) for delivering data to the client, again.

    The method :meth:`to_bytes` transforms this into a bytes response.

    :ivar code: The response code.
    :ivar cookies: The cookies to send down with the response.
    :ivar body: The string or bytes body of the request.
    :ivar headers: A dict of headers for the response.
    :ivar request: The request object this Response is handling.
            This is automatically set inside Kyoukai.
    """

    def __init__(self, code: int, body: typing.Union[str, bytes], headers: dict = None):
        """
        Create a new response.
        """
        self.code = code
        self.cookies = SimpleCookie()
        self.body = body
        self.headers = IOrderedDict(headers) if headers else IOrderedDict()

        self._should_gzip = False
        self._is_head = False

        self.request = None

    @property
    def gzip(self):
        """
        :return: If the request is to be gzip compressed.

        Note: This will always return False on a newly created Response, unless the ``request`` instance variable is
        set on the Response.
        """
        if self.request and self.request.fully_parsed:
            return ('gzip' in self.request.headers.get("Accept-Encoding", "")) and self._should_gzip
        else:
            return False
            # return self._should_gzip

    @gzip.setter
    def gzip(self, value):
        self._should_gzip = value

    # Utility functions.

    def get_compressed_body(self, body: bytes) -> bytes:
        """
        Returns the compressed body, if gzip is enabled.

        Otherwise, returns the normal body.
        :param body: The body to compress.
        :return: The compressed body, or just the body.
        """
        if self.gzip:
            return gzip.compress(body, 5)
        else:
            return body

    def get_response_http_version(self):
        """
        Gets what HTTP version the response should use.
        """
        if self.request and self.request.fully_parsed:
            return '.'.join(map(str, self.request.version))
        else:
            return "1.0"

    def _mimetype(self, body):
        """
        Calculates the mime type of the response, using libmagic.

        This is an **internal method**.
        """
        if _has_magic:
            # libmagic is unreliable when the body is str.
            # so encode it to make sure it returns correctly
            if isinstance(self.body, str):
                mime = magic.from_buffer(body.encode(), mime=True)
            else:
                mime = magic.from_buffer(body, mime=True)
            if mime:
                return mime.decode() if isinstance(mime, bytes) else mime
            else:
                return "empty"
        else:
            return "text/plain"

    def _recalculate_headers(self):
        """
        Override certain headers, like Content-Size.

        This is an **internal method**.
        """
        if not self._is_head:
            # The +2 is for the \r\n at the end.
            self.headers["Content-Length"] = len(self.body) + 2
        if 'Content-Type' not in self.headers:
            self.headers["Content-Type"] = self._mimetype(self.body) or "text/plain"

        # If it's gzip enabled, add the gzip header.
        if self.gzip:
            self.headers["Content-Encoding"] = "gzip"

        # Set cookies.
        self.headers["Date"] = formatdate()
        self.headers["Server"] = "Kyoukai/{} (see https://github.com/SunDwarf/Kyoukai)".format(util.VERSION)
        self.headers["X-Powered-By"] = "Kyoukai"

    def to_bytes(self):
        """
        Serialize a Response into :class:`bytes` to return and send to the client.

        :return: The encoded data for the response.
        """
        if self.request and self.request.fully_parsed:
            if self.request.method.lower() == "head":
                self._is_head = True

        version = self.get_response_http_version()

        if isinstance(self.body, str):
            self.body = self.body.encode()
        elif isinstance(self.body, bytes):
            pass
        else:
            self.body = str(self.body).encode()

        if self.gzip:
            if 'Content-Type' not in self.headers:
                self.headers["Content-Type"] = self._mimetype(self.body)
            self.body = self.get_compressed_body(self.body)

        # Re-calculate headers to update everything as appropriate.
        self._recalculate_headers()

        fmt = "HTTP/{version} {code} {msg}\r\n{headers}{cookies}\r\n"
        headers_fmt = ""
        # Calculate headers
        for name, val in self.headers.items():
            headers_fmt += "{}: {}\r\n".format(name, val)

        # Get the HTTP code.
        code = http.HTTPStatus(self.code).name.replace("_", " ")

        built = fmt.format(code=self.code, msg=code, headers=headers_fmt,
                           cookies=(self.cookies.output() + "\r\n") if len(self.cookies) else "",
                           version=version)

        # Encode the built string so far.
        built = built.encode()

        # Append the body, plus the terminator.
        built += self.body + b"\r\n"

        return built

    @classmethod
    def redirect(cls, location, code=302):
        """
        Creates a new Response that redirects to a specific location.

        :param location: The location to redirect to.
        :param code: The code (usually a 301 or 302) to add to the response.
        :return: A new :class:`Response` for the redirect.
        """
        # https://github.com/pallets/werkzeug/blob/master/werkzeug/utils.py#L373
        # response body used from Werkzeug

        res = cls(
            code=code,
            body=
            '<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 3.2 Final//EN">\n'
            '<title>Redirecting...</title>\n'
            '<h1>Redirecting...</h1>\n'
            '<p>You should be redirected automatically to target URL: '
            '<a href="{location}">{location}</a>.  If not click the link.'.format(location=location),
            headers={"Location": location}
        )
        return res


# Alias for ease of usage
redirect = Response.redirect



import mako.lookup

# Create a lookup that uses /templates and /tmp/mako_modules to store cached data.
kyokai_templates = mako.lookup.TemplateLookup(directories=['templates'], module_directory='/tmp/mako_modules')


def render(filename, **kwargs):
    """
    Render a file using Mako.

    Passes all of the keyword arguments to the template renderer.
    """
    template = kyokai_templates.get_template(filename)
    return template.render(**kwargs)


"""
Converters.

Defines how to convert arguments in a Route via the signature.
"""
import inspect
import typing

from kyoukai.exc import HTTPException
from kyoukai.context import HTTPRequestContext

_converters = {
    str: lambda ctx, item: item,
    int: lambda ctx, item: int(item)
}


def add_converter(type_: type, cb: typing.Callable[[HTTPRequestContext, typing.Any], typing.Any]):
    """
    Adds a converter to the conversions list/
    :param type_: The type to use as the annotation param.
    :param cb: A callable.
            Takes two parameters: The HTTPRequestContext, and the item to convert.

            This callable should raise a TypeError or a ValueError on failing to convert, at which point a 400 error
            will be raised.
            Anything else will cause a normal 500 error.
    """
    if not callable(cb):
        raise TypeError("cb should be callable")

    _converters[type_] = cb


def convert_args(ctx, coro, *args, bound=False):
    """
    Converts a the arguments of a function using it's signature.

    Will ignore `self` if bound is True.

    :param coro: The coroutine function to inspect for the signature.
    :param args: The arguments that are to be passed into the function.
            The first one should be the HTTPRequestContext; this is ignored.
    :param bound: If this route is bound to a View.
            Setting this will ignore the first parameter of the signature.
    """
    signature = inspect.signature(coro)
    params = signature.parameters

    if len(args) != len(params):
        raise IndexError("Arguments passed in were not the same length as {}'s function signature".format(coro))

    new_args = []

    for num, (name, value) in enumerate(params.items()):
        # If bound, just ignore the `self` param.
        if bound and num == 0:
            new_args.append(args[0])
            continue

        item = args[num]
        # Skip over the HTTPRequestContext.
        if isinstance(item, HTTPRequestContext):
            new_args.append(item)
            continue

        # Extract the annotation from the parameter.
        assert isinstance(value, inspect.Parameter)
        type_ = value.annotation
        if type_ not in _converters:
            # Just add the argument, without converting.
            new_args.append(item)
        else:
            _converter = _converters[type_]
            # Convert the arg.
            try:
                new_args.append(_converter(item))
            except (TypeError, ValueError) as e:
                # Raise a bad request error.
                raise HTTPException(400) from e

    return new_args



"""
The Base renderer is a simple ABC which children renderers inherit from and override.
"""
import abc


class Renderer:
    @abc.abstractmethod
    def get_template(self, template_name: str):
        """
        Gets a template from the rendering environment.

        :param template_name: The path to the template.
        :return: The Template object for the respective Renderer.
        """

    @abc.abstractmethod
    def render(self, template_name: str, **kwargs) -> str:
        """
        Renders a template out into a HTML string.

        :param template_name: The name of the template to use.
        :param kwargs: Keyword arguments to pass into the renderer.
        :return: A new :class:`str` which is the output of the renderer.
        """



"""
Kyoukai exceptions.
"""


class HTTPException(Exception):
    """
    A basic HTTP error.

    This is used to quickly escape out of a function, by raising the appropriate HTTP Exception.

    .. code:: python

        if not some_condition:
            raise HTTPException(404)
    """
    def __init__(self, errcode, msg=None,
                 route=None):
        self.code = errcode

        self.msg = msg

        self.route = route

    def __repr__(self):
        return "HTTP {} {}".format(self.code, self.msg)



from .app import Kyoukai
from .request import Request
from .response import Response
from .context import HTTPRequestContext
from .blueprints import Blueprint
from .route import Route
from .views import View
from .asphalt import KyoukaiComponent
from .exc import HTTPException
from .util import VERSION, VERSIONT, static_filename
from .protocol import KyoukaiProtocol



"""
Views are like grouped together routes - but not Blueprints. They allow creating classes for routes, and storing
data, etc inside the class.
"""
from kyoukai.route import Route


def route(regex, methods: list = None, route_cls: type=Route):
    """
    Create a "bound" route.

    This should then be updated by ViewMeta with the correct self param, and added to the blueprint with the
    Blueprint.bind_view() method.

    This route is meant to be placed in a class that inherits from :class:`View`.

    .. versionchanged:: 1.8

        This method now accepts an optional ``route_cls`` which designates which Route class to wrap the coroutines in.

    :param regex: The regex to use for routing.

    :param methods: The methods that are allowed for this route.

    :param route_cls: The Route class to wrap this Route in.

    :return: The :class:`Route` that wraps this coroutine.
    """
    if not methods:
        methods = ["GET"]

    def _route_inner(coro):
        new_route = route_cls(None, regex, methods, bound=True)
        new_route.create(coro)
        return new_route

    return _route_inner


class ViewMeta(type):
    """
    Defines a way to automatically update routes.

    Internal metaclass.
    """
    __registry = {}

    def __call__(cls, *args, **kwargs):
        """
        Establish a singleton class, and update Route objects as appropriate.
        """
        name = cls.__name__
        if name in ViewMeta.__registry:
            return ViewMeta.__registry[name]

        new_class = super().__call__(*args, **kwargs)

        # Give the new_class a `_binded` property.
        # Used for the Blueprint.
        new_class._binded = False

        # Loop over the cls' dict.
        for item in cls.__dict__.values():
            if isinstance(item, Route):
                # New Route item.
                if item._bound:
                    item.self = new_class

        ViewMeta.__registry[name] = new_class

        return new_class

    def get_routes(cls):
        """
        :return: A list of Route objects for the specified View.
        """
        return [i for i in cls.__dict__.values() if isinstance(i, Route)]


class View(metaclass=ViewMeta):
    """
    Stub class to allow inheriting from easily without having to do metaclass=ViewMeta.
    """



"""
Stores HTTPRequestContext
"""
from asphalt.core import Context
from typeguard import check_argument_types

import kyoukai


class HTTPRequestContext(Context):
    """
    Sub-class of context used for HTTP requests.
    """

    cfg = {}

    def __init__(self, request, app, parent: Context):
        assert check_argument_types()
        super().__init__(parent=parent)
        self._request = request

        self._app = app

    @property
    def request(self) -> 'kyoukai.Request':
        return self._request

    @request.setter
    def request(self, req: 'kyoukai.Request'):
        self._request = req

    @property
    def app(self) -> 'kyoukai.Kyoukai':
        return self._app



"""
Testing application for Kyoukai.
"""
import asyncio
import collections
import typing

from asphalt.core import Context

from kyoukai import HTTPRequestContext
from kyoukai import Request
from kyoukai import Response
from kyoukai.app import Kyoukai


class TestProtocol:
    """
    A fake test protocol, that has a mock close and send method to emulate a real transport for the app.
    """
    def __init__(self, app: 'TestKyk'):
        self.app = app

        # Make a fake lock, for the Request lock.
        self.lock = asyncio.Lock()

    def handle_resp(self, response: Response):
        self.app._responses.put_nowait(response)

    def close(self):
        pass


class TestingKyk(Kyoukai):
    """
    A test application for Kyoukai.

    This has a few notable differences from a regular app:

     - The :meth:`TestKyk.feed_request` method that feeds a new request to be processed.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._testing_protocol = TestProtocol(self)

        self._responses = asyncio.Queue()

    async def feed_request(self, data: typing.Union[str, Request]) -> Response:
        """
        Feeds a new Request object to the app.

        :param data: Either a string containing request data, or a raw Request object.
        :return: The Response object produced from `delegate_request`.
        """
        if isinstance(data, str):
            data = data.replace("\n", "\r\n")
            if not data.endswith("\r\n\r\n"):
                data += "\r\n"
            request = Request.from_data(data.encode(), "127.0.0.1")
        else:
            request = data

        # Create a new HTTPRequestContext.
        ctx = HTTPRequestContext(request, self, Context())

        # Delegate the request.
        await self.delegate_request(self._testing_protocol, ctx)
        # This should return None, as the queue was updated during delegate_request.
        return self._responses.get_nowait()

    def clean(self):
        """
        Cleans up the state of the responses Queue after a request.
        """
        self._responses._queue = collections.deque()





"""
Jinja2 renderer for templates.
"""
from jinja2.environment import Environment, Template
# Default loader is the FileSystemLoader, contained inside a ChoiceLoader.
from jinja2.loaders import FileSystemLoader, ChoiceLoader, BaseLoader

from kyoukai.renderers.base import Renderer


class Jinja2Renderer(Renderer):
    """
    A renderer that uses the Jinja2 Rendering Engine.

    :param loader: The Loader to use for the engine to load templates from.
    """
    def __init__(self, loader: BaseLoader=None):
        self._loader = ChoiceLoader(loaders=[])

        if loader is None:
            # Use a default FileSystemLoader.
            self._loader.loaders.append(FileSystemLoader(searchpath="templates"))

        self._environment = None

    @property
    def loader(self):
        return self._loader

    @property
    def environment(self) -> Environment:
        if self._environment is None:
            self._environment = Environment(loader=self._loader)
        return self._environment

    @environment.setter
    def environment(self, env: Environment):
        self._environment = env

    def add_loader(self, loader: BaseLoader):
        """
        Adds a loader to the ChoiceLoader contained within.

        :param loader: The Loader to add.
        """
        self.loader.loaders.append(loader)

    def get_template(self, template_name: str) -> Template:
        """
        Gets a :class:`jinja2.environment.Template` from the Environment.

        :param template_name: The template to load.
        :return: The new template object.
        """
        return self.environment.get_template(template_name)

    def render(self, template_name: str, **kwargs) -> str:
        """
        Renders a template using Jinja2.

        :param template_name: The name of the template to render.
        :param kwargs: Keyword arguments to pass to the template as a dict.
        :return: A new str containing the rendered HTML of the template.
        """
        template = self.get_template(template_name)
        return template.render(**kwargs)





import asyncio
import logging

from kyoukai.exc import HTTPException
from kyoukai.request import Request
from kyoukai.response import Response
from kyoukai.context import HTTPRequestContext


# This probably can't be reasonably tested.
# And besides, the TestProtocol emulates this as well.
class KyoukaiProtocol(asyncio.Protocol):  # pragma: no cover
    """
    The Kyoukai protocol.
    """

    def __init__(self, app, parent_context):
        self.app = app
        self._transport = None
        self.ip = None
        self.client_port = None

        self.logger = logging.getLogger("Kyokai")

        self.loop = asyncio.get_event_loop()

        # Asphalt contexts
        self.parent_context = parent_context

        self.buffer = b""

        # Request lock.
        # This ensures that requests are processed serially, and responded to in the correct order, as the lock is
        # released after processing a request completely.
        self.lock = asyncio.Lock()

    def connection_made(self, transport: asyncio.Transport):
        """
        Called when a connection is made, and is used to store the connection data.
        """
        self.ip, self.client_port = transport.get_extra_info("peername")
        self._transport = transport

        self.logger.debug("Recieved connection from {}:{}".format(*transport.get_extra_info("peername")))

    def handle_resp(self, res: Response):
        """
        Handle a response.

        :param res: The response to write into the transport.
        """
        data = res.to_bytes()
        self._transport.write(data)

    def data_received(self, data: bytes):
        """
        Create a new Request, and delegate Kyokai to process it.
        """
        self.logger.debug(
            "Recieved {} bytes of data from client {}:{}, feeding.".format(
                len(data), *self._transport.get_extra_info("peername")
            )
        )

        # Delegate as response.
        self.logger.debug("Delegating response for client {}:{}.".format(*self._transport.get_extra_info("peername")))
        # Create a request
        self.buffer += data
        req = self.app.request_cls()
        ctx = HTTPRequestContext(req, self.app, self.parent_context)
        try:
            req.parse(self.buffer, self.ip)
        except HTTPException as e:
            # Delegate the HTTP exception, probably a 400.
            self.app.log_request(e, code=e.code)
            self.loop.create_task(self.app.handle_http_error(e, self, ctx))
        else:
            if req.fully_parsed:
                # Reset buffer.
                self.logger.debug("Request for `{}` fully parsed, passing.".format(req.path))
                self.buffer = b""
                self.loop.create_task(self.app.delegate_request(self, ctx))
            else:
                # Continue.
                return

    def close(self):
        # Close the protocol and empty the locks.
        for waiter in self.lock._waiters:
            self.logger.warn("Cancelling waiter {} for {}...".format(waiter, self.lock))
            waiter.cancel()
        self._transport.close()



"""
Kyōkai blueprints are simply groups of routes.

They're a simpler way of grouping your routes together instead of having to import your app object manually all of
the time.
"""
import collections
import logging

import re

from kyoukai.exc import HTTPException
from kyoukai.route import Route
from kyoukai.views import View


class Blueprint(object):
    """
    A Blueprint is a container for routes.

    Blueprints have 'parent' blueprints - they inherit error handlers and hooks from them. The root blueprint has no
    parent, so it does not inherit from anything.
    Note that if a Blueprint that is not the root blueprint has a parent value of None, it is automatically set to
    inherit the root blueprint of the app.

    Blueprints allow routes like normal application objects - in fact, the application object's route function is
    implemented with an underlying root blueprint.

    .. code:: python

        bp = Blueprint("test_bp")

        @bp.route("/hello/world")
        async def hello_world(ctx: HTTPRequestContext):
            return "Hello, world"!

    Per-blueprint error handlers can also be added:

    .. code:: python

        @bp.errorhandler(500)
        async def handle_500(ctx: HTTPRequestContext, e: Exception):
            async with threadpool():
                ctx.dbsession.rollback()

    :param name: The name identifier of the blueprint.
    :param parent: The parent blueprint.
        Children blueprint inherit routes from the parent, accessible via ``Parent.routes``.
        They are also used for searching in routes when a blueprint is checked.

        If this is None, then it will be automatically set when the blueprint is added as a child to a parent
        blueprint.

    :param url_prefix: The prefix to automatically add to the start of each route.

    :cvar route_cls: The Route class to use for wrapping new Routes.
    """

    route_cls = Route

    def __init__(self, name: str, parent: 'Blueprint' = None,
                 url_prefix: str = ""):
        self._prefix = url_prefix
        self._name = name

        self.routes = []

        self.errhandlers = {}

        # The parent of this blueprint.
        self._parent = parent
        if self._parent is not None:
            # Add ourselves as a child.
            self._parent.add_child(self)

        # The children of this blueprint.
        self._children = {}

        self._request_hooks = collections.defaultdict(lambda *args, **kwargs: collections.OrderedDict())

        self.logger = logging.getLogger("Kyoukai.Blueprint." + self._name if self._name else "root")

    def bind_view(self, view: View, *args, **kwargs):
        """
        Binds a view class to a Blueprint.

        This takes the *class*, not the instance, as a param.

        It also takes args and keyword args to instantiate the class with.
        """
        # Create a new instance of the class.
        new_view = view(*args, **kwargs)
        # Don't bind if it's already binded.
        if new_view._binded:
            return
        for route in view.get_routes():
            route.bp = self
            route.matcher = re.compile(self.prefix + route._match_str)
            self.routes.append(route)

        # Add before_request and after_request from the new_view.
        def __wrapper(func):
            async def ___internal_wrapper(*args, **kwargs):
                await func(*args, **kwargs)

            return ___internal_wrapper

        if hasattr(new_view, "before_request"):
            # Create wrappers for the new views.
            nf = __wrapper(new_view.before_request)
            nf.__name__ = new_view.__class__.__name__ + ".before_request"
            self.before_request(nf)

        if hasattr(new_view, "after_request"):
            # Create wrappers for the new view.
            nf = __wrapper(new_view.after_request)
            nf.__name__ = new_view.__class__.__name__ + ".after_request"
            self.after_request(nf)

        self.logger.info("Bound view {}".format(view.__name__))

    def __repr__(self):
        return "<Blueprint '{}' with {} routes>".format(self._name, len(self.routes))

    def add_child(self, blueprint: 'Blueprint'):
        """
        Add a child Blueprint to the current blueprint.

        .. warning::
            This will override the parent of the blueprint, replacing it with this one.

        :param blueprint: The child blueprint.
        """
        self._children[blueprint._name] = blueprint
        self.logger.info("Registered Blueprint {} with {} new routes".format(blueprint._name, len(blueprint.routes)))
        blueprint.parent = self

    def before_request(self, coro):
        """
        Set a coroutine to run as before the request.

        This coroutine should take in the HTTPRequestContext, and return a new HTTPRequestContext.
        """
        # Unlike before, don't use Route objects.
        # Just append it to the route handlers.
        self._request_hooks["pre"][coro.__name__] = coro
        return coro

    def after_request(self, coro):
        """
        Set a coroutine to run after the request.

        This coroutine should take in a :class:`Response`, and return a :class:`Response`.
        """
        self._request_hooks["post"][coro.__name__] = coro
        return coro

    def get_pre_hooks(self, ctx):
        """
        Get the pre-request hooks in a list.

        This goes from top-level blueprint to bottom-level blueprint in terms of order.

        :param ctx: The context of the request.
        """
        if self.parent is not None:
            bps = self.parent.get_pre_hooks(ctx)
        else:
            bps = []

        # Return the list of our hooks, merged with the other list.
        return bps + list(self._request_hooks["pre"].values())

    def get_post_hooks(self, ctx):
        """
        Get the post-request hooks in a list.

        This goes from top-level blueprint to bottom-level blueprint in terms of order.

        :param ctx: The context of the request.
        """
        if self.parent is not None:
            bps = self.parent.get_post_hooks(ctx)
        else:
            bps = []

        # Return the list of our hooks, merged with the other list.
        return bps + list(self._request_hooks["post"].values())

    def match(self, route: str, method: str):
        """
        Match a route.

        This will search down our routes, and then down our children's routes, to see if we can find a match for the
        specified route.

        :param route: The route to match, e.g ``/abc/def``.
        :param method: The method of the route.

        :raises: A :class:`kyoukai.exc.HTTPException` with code 415 if the method is not valid.
        :returns: The :class:`Route` if the route was matched, or None.
        """
        matches = []
        for route_obb in self.routes:
            assert isinstance(route_obb, Route)
            matched = route_obb.match(route, method)
            if matched:
                matches.append(matched)

        # Search through the children
        for child in self.children.values():
            matched = child.match(route, method)
            if matched:
                matches.append(matched)

        if not matches:
            return None
        else:
            # Loop through each route, and check the method.
            # If no method matches, then raise the HTTPException. Otherwise, return the route.
            # This allows for multiple routes with the same method.
            for route in matches:
                if route.kyokai_method_allowed(method):
                    return route
            else:
                # This is called when no return successfully hit.
                raise HTTPException(405)

    @property
    def parent(self) -> 'Blueprint':
        """
        :returns: The parent of this blueprint.
        """
        return self._parent

    @parent.setter
    def parent(self, bp: 'Blueprint'):
        """
        Sets the parent blueprint.
        """
        self._parent = bp

    @property
    def children(self):
        """
        :returns: A :class:`list` of the children of this blueprint.
        """
        return self._children

    @property
    def prefix(self):
        """
        Calculates the prefix using the parent blueprints.
        :return: The full prefix, including the parent prefix.
        """
        if self.parent is None:
            return self._prefix
        return self.parent.prefix + self._prefix

    def wrap_route(self, regex, coroutine, *, methods: list = None, run_hooks=True):
        """
        Wraps a route in a :class:`Route` object.

        The class that this returns can be configured with the ``route_cls`` attribute of the Blueprint. it is
        automatically called with the specified regular expression.

        :param regex: The regular expression to match the path to. This uses standard Python :mod:`re` syntax.
                Group matches are automatically extracted from the regex, and passed as arguments.

        :param methods: The list of allowed methods, e.g ["GET", "POST"].
                You can check the method with `request.method`.

        :param coroutine: The coroutine handler to take in, which

        :param run_hooks: If pre and post request hooks are ran.

        :return: The new :class:`Route` object.
        """
        if not methods:
            methods = ["GET"]

        regex = self.prefix + regex
        r = self.route_cls(self, regex, methods, run_hooks=run_hooks)
        r.create(coroutine)

        return r

    def add_route(self, route: Route):
        """
        Adds a route object to the routing table.

        .. note::

            This will set the blueprint of the specified route to ourselves, overwriting any other routes.

        :param route: The route object to add.
        :return: The Route.
        """
        route.bp = self
        self.routes.append(route)

        return route

    def route(self, regex, *, methods: list = None, run_hooks=True):
        """
        Convenience decorator to create a new route.

        This is equivalent to:

        .. code:: python

            route = bp.wrap_route(regex, callable, methods, run_hooks)
            bp.add_route(route)


        :param regex: The regular expression to match the path to. This uses standard Python :mod:`re` syntax.
                Group matches are automatically extracted from the regex, and passed as arguments.

        :param methods: The list of allowed methods, e.g ["GET", "POST"].
                You can check the method with `request.method`.

        :param run_hooks: Should the pre and post request hooks run automatically?
                This is set to True by default.
        """

        def _add_route_inner(coro):
            route = self.wrap_route(regex, coro, methods=methods, run_hooks=run_hooks)
            self.add_route(route)
            return route

        return _add_route_inner

    def get_errorhandler(self, code: int) -> Route:
        """
        Not to be used by the user/dev - internal function used to traverse down the parent's error handlers.
        """
        err_handler = self.errhandlers.get(code)

        if err_handler:
            return err_handler

        if self.parent is None:
            # We can't get the parent's error handler, oh well.
            return None

        return self.parent.get_errorhandler(code)

    def add_errorhandler(self, code: int, err_handler: Route):
        """
        Adds an error handler to the dictionary of error handlers for this route.

        :param code: The error code that this error handler should handle.

        :param err_handler: A :class:`Route` object that handles the error.
                Error handlers are just modified routes, so the use of a Route object here is correct.

        :return: The original Route, but with the ``bp`` attribute set to this Blueprint.
        """
        err_handler.bp = self
        self.errhandlers[code] = err_handler

        return err_handler

    def errorhandler(self, code: int, run_hooks=False):
        """
        Convenience decorator to add an error handler.

        This is equivalent to:

        .. code:: python

            route = bp.wrap_route("", coro, methods=[], run_hooks=False)
            bp.add_errorhandler(code, route)
        """

        def _add_route_inner(coro):
            route = self.wrap_route("", coro, methods=[], run_hooks=run_hooks)
            self.add_errorhandler(code, route)
            return route

        return _add_route_inner



"""
Mako renderer for templates.
"""
import posixpath

from mako import exceptions
from mako.util import to_list
from mako.lookup import TemplateLookup
from mako.template import Template

from kyoukai.renderers.base import Renderer


class MakoRenderer(Renderer):
    """
    A renderer that uses the Mako templating engine.

    :ivar environment: The global environment for this renderer.
        This is passed into the template renderer every time.
    """

    def __init__(self, lookup_directories: list = None):
        # Template lookup, which is accessed via a setting property.
        self._lookup = None

        # Lookup path.
        if not lookup_directories:
            self._lookup_directories = ['templates']
        else:
            self._lookup_directories = lookup_directories

        # Global environment.
        self.environment = {}

    @property
    def template_lookup(self) -> TemplateLookup:
        """
        :return: The :class:`mako.lookup.TemplateLookup` for this renderer.
        """
        if self._lookup is None:
            self._lookup = TemplateLookup(directories=self._lookup_directories,
                                          module_directory='/tmp/mako_modules')
        return self._lookup

    @template_lookup.setter
    def template_lookup(self, val):
        self._lookup = val

    def add_template_directories(self, *directories):
        """
        Adds a directories to the lookup paths for the templates.
        :param directories: The directories to add.
        """
        lookup = self.template_lookup
        # Process the directories to be "mako compatible."
        dirs = [posixpath.normpath(d) for d in to_list(directories, ())]
        lookup.directories += dirs

    def get_template(self, filename) -> Template:
        """
        Gets a :class:`mako.template.Template` object from the specified filename.

        :param filename: The filename of the template to retrieve.
        :return: The template returned from the disk.
        """
        # Find the template using the lookup.
        return self.template_lookup.get_template(filename)

    def render(self, filename, handle_exceptions=False, **kwargs) -> str:
        """
        Render a new template using Mako.

        :param filename: The filename of the template to lookup and render.
        :param handle_exceptions: Should a rendered error template be returned in case of an exception?
        :param kwargs: The keyword arguments to pass to the renderer.
        :return: The rendered template, in string format.
        """
        template = self.get_template(filename)
        try:
            to_pass = {**self.environment, **kwargs}
            return template.render(**to_pass)
        except Exception as e:
            if not handle_exceptions:
                raise
            else:
                # Render the error template.
                return exceptions.text_error_template().render()



"""
The current set of test suite data used for Kyoukai.
"""
import json

from kyoukai import HTTPRequestContext
from kyoukai import Response
from kyoukai.testing.testapp import TestingKyk

kyk = TestingKyk("")


@kyk.route("/")
async def root(ctx: HTTPRequestContext):
    return "OK"


@kyk.route("/headers")
async def headers(ctx: HTTPRequestContext):
    r = Response(body="OK", code=200)
    r.headers = ctx.request.headers
    return r


@kyk.route("/params")
async def url_params(ctx: HTTPRequestContext):
    r = Response(body="", code=204)
    for param in ctx.request.args:
        r.headers[param] = ctx.request.args[param]

    return r


@kyk.route("/json", methods=["POST"])
async def js_data(ctx: HTTPRequestContext):
    body = {}
    for key, value in ctx.request.form.items():
        body[value] = key

    return json.dumps(body), 204



"""
A Kyoukai app is the core container of a web application based upon the framework.
"""

import asyncio
import inspect
import io
import mimetypes
import os
import logging
import traceback
import typing

try:
    import magic
except (ImportError, OSError):
    _has_magic = False
else:
    _has_magic = True
from asphalt.core import Context

from asphalt.core.runner import run_application
from typeguard import check_argument_types

from kyoukai.debugger import KyoukaiDebugger

from kyoukai.blueprints import Blueprint
from kyoukai.context import HTTPRequestContext
from kyoukai.util import static_filename, wrap_response
from kyoukai.exc import HTTPException

from kyoukai.response import Response
from kyoukai.request import Request

from kyoukai.route import Route
from kyoukai.views import View

from kyoukai.renderers.base import Renderer
from kyoukai.renderers import mako_renderer
try:
    from kyoukai.renderers import jinja_renderer
    _has_jinja2 = True
except ImportError:
    jinja_renderer = None
    _has_jinja2 = False


class Kyoukai(object):
    """
    A Kyoukai app.

    This is the core component to your web application based on Kyoukai.

    :param name:
        The name of the app. This is passed into the root blueprint as the name, which shows up in exceptions.
    :type name: :class:`str`

    :cvar request_cls: The class to use to create Requests.
    :cvar response_cls: The class to use to create wrapped Responses.
    """
    request_cls = Request
    response_cls = Response

    def __init__(self, name: str, **kwargs):
        """
        Create a new app.
        """

        self.name = name
        self.loop = asyncio.get_event_loop()

        self.logger = logging.getLogger("Kyoukai")

        self.error_handlers = {}

        # Define the config.
        self.config = kwargs

        # Define the "root" blueprint, which is used for @app.request.
        self._root_bp = Blueprint(name, None)

        self.debug = kwargs.get("debug", False)
        self._debugger = KyoukaiDebugger(self)

        # Define the component here so it can be checked easily.
        self.component = None

        # On startup function.
        self._on_startup = lambda app: None

        # Define the renderer.
        render = kwargs.get("renderer", "mako")
        if render == "mako":
            self._renderer = mako_renderer.MakoRenderer(kwargs.get("template_directory"))
        elif render == "jinja2":
            if not _has_jinja2:
                raise ImportError("Jinja2 is not installed; cannot be used as a renderer")
            else:
                self._renderer = jinja_renderer.Jinja2Renderer(kwargs.get("loader"))

    def reconfigure(self, **cfg: dict):
        """
        Reconfigures the app using the parameters passed in from the config.

        This is used, for example, with Asphalt's config and the KyoukaiComponent's consume rest behaviour when
        dealing with keyword arguments.

        :param cfg: A dictionary of config values to reconfigure the application with.
        :return: The new configuration.
        """
        self.config = {**self.config, **cfg}
        return self.config

    @property
    def renderer(self) -> Renderer:
        """
        :return: The current :class:`kyoukai.renderers.Renderer` configured for this application.
        """
        return self._renderer

    @property
    def root(self):
        """
        This property is a way to access the root blueprint.

        The root blueprint is the parent of all registered blueprints in the application. It can be used to directly
        route onto the top-level, however it is recommended to just use :meth:`Kyoukai.route` and similar to run onto
        the root blueprint.
        """

        return self._root_bp

    def register_blueprint(self, bp: Blueprint):
        """
        Registers a blueprint as a sub-blueprint to the root blueprint.

        :return: The modified Blueprint with the correct parent.
        """
        assert check_argument_types()
        self._root_bp.add_child(bp)
        if bp.parent is None:
            bp.parent = self._root_bp
        return bp

    def render(self, filename: str, **kwargs) -> str:
        """
        Render a template using the currently loaded rendering engine.

        :param filename: The filename of the template to load and render.
        :type filename: str

        :param kwargs: Additional arguments to pass to the template renderer.
                These are directly passed to :meth:`mako.template.Template.render` to render the template.

        :return: A :class:`str` containing the rendered information.
        """
        return self.renderer.render(filename, **kwargs)

    def render_template(self, filename: str, code=200, **kwargs) -> Response:
        """
        Render a template using the currently loaded rendering engine.

        Unlike :meth:`Kyoukai.render`, this returns a :class:`Response` object.

        :param filename: The filename of the template to load and render.
        :type filename: str

        :param code: The response code to add into the Response.

        :param kwargs: Additional arguments to pass to the template renderer.
                These are directly passed to :meth:`mako.template.Template.render` to render the template.

        :return: A :class:`Response` object with the rendered template.
        """
        data = self.render(filename, **kwargs)
        # Wrap it in a response.
        return Response(code, data, headers={"Content-Type": "text/html"})

    def get_static_path(self, filename: str) -> str:
        """
        Gets a sanitized static path.

        This will use the current working directory and the `static_dir` defined by the configuration, or `static` if
        that is not specified, to build the full path.

        :param filename: The filename to look up.
        :return: A :class:`str` with the fully built path.
        """
        return os.path.join(os.getcwd(), self.config.get("static_dir", "static"), static_filename(filename))

    def get_static_file(self, filename: str) -> io.BufferedIOBase:
        """
        Opens and returns a static file from the path.

        Internally, this uses :meth:`Kyoukai.get_static_path` to sanitize the path.

        :param filename: The filename to load.
        :return: A file object opened in binary mode, attached to the file.
        :rtype: :class:`io.BufferedIOBase`
        """
        fname = self.get_static_path(filename)
        if not os.path.exists(fname):
            return None
        else:
            return open(os.path.join(os.getcwd(), self.config.get("static_dir", "static"), fname), 'rb')

    def get_static(self, filename: str) -> Response:
        """
        Gets a file, using static, but returns a Response instead of the file handle.
        """
        content = self.get_static_file(filename)
        if not content:
            raise HTTPException(404)

        with content:
            path = self.get_static_path(filename)
            mimetype = mimetypes.guess_type(path)[0]
            if not mimetype:
                if _has_magic:
                    mimetype = magic.from_file(path, mime=True).decode()
                else:
                    mimetype = "application/octet-stream"
            return Response(200, body=content.read(), headers={"Content-Type": mimetype})

    def _match_route(self, path, meth) -> Route:
        """
        Match a route, based on the regular expression of the route.

        :return: The :class:`kyoukai.Route` of the route that matches, or None if no route matches.
        :raises: :class:`kyoukai.exc.HTTPException` if the route matches, but the method is not allowed.
        """
        return self._root_bp.match(path, meth)

    def route(self, regex, *, methods: list = None, run_hooks=True):
        """
        Convenience decorator to create a new route.

        This is equivalent to:

        .. code:: python

            route = bp.wrap_route(regex, callable, methods, run_hooks)
            bp.add_route(route)

        .. note::

            This function is a shortcut to ``app.root.route(*args, **kwargs)``.

        :param regex: The regular expression to match the path to. This uses standard Python :mod:`re` syntax.
                Group matches are automatically extracted from the regex, and passed as arguments.

        :param methods: The list of allowed methods, e.g ["GET", "POST"].
                You can check the method with `request.method`.

        :param run_hooks: Should the pre and post request hooks run automatically?
                This is set to True by default.
        """
        return self._root_bp.route(regex, methods=methods, run_hooks=run_hooks)

    def errorhandler(self, code: int):
        """
        Convenience decorator to add an error handler.

        This is equivalent to:

        .. code:: python

            route = bp.wrap_route("", coro, methods=[], run_hooks=False)
            bp.add_errorhandler(code, route)

        .. note::

            This function is a shortcut to ``app.root.errorhandler(code)``.
        """
        return self._root_bp.errorhandler(code)

    def log_request(self, ctx: HTTPRequestContext, code: int = 200):
        """
        Logs a request to the logger.

        This is an **internal** method and should not be used by user code.

        :param ctx: The :class:`kyoukai.context.HTTPRequestContext` to log from.
        :param code: The error code to log. Defaults to 200.
        """
        try:
            route = ctx.request.path
        except AttributeError:
            # *really* bad request - ignore it.
            return
        self.logger.info("HTTP/{} {} {} - {}".format(ctx.request.sversion, ctx.request.method, route, code))

    def before_request(self, func):
        """
        Set a coroutine to run as before the request.

        This coroutine should take in the HTTPRequestContext, and return a new HTTPRequestContext.

        .. note::

            This function is a shortcut to ``app.root.before_request(func)``.
        """
        return self._root_bp.before_request(func)

    def after_request(self, func):
        """
        Set a coroutine to run after the request.

        This coroutine should take in a :class:`Response`, and return a :class:`Response`.

        .. note::

            This function is a shortcut to ``app.root.after_request(func)``.
        """

        return self._root_bp.after_request(func)

    def bind_view(self, view: View):
        """
        Binds a view class to a Blueprint.

        This takes the *class*, not the instance, as a param.

        It also takes args and keyword args to instantiate the class with.

        .. note::

            This function is a shortcut to ``app.root.bind_view(view)``.
        """
        self._root_bp.bind_view(view)

    async def call_on_startup(self):
        """
        Calls the on_startup handler.

        This should not be called, unless you wish to call the startup function again (?).

        .. versionchanged:: 1.8

            This now passes the app object.
        """
        try:
            item = self._on_startup(self)
        except:
            # Only print the bottom most traceback.
            tb = traceback.format_exc()
            self.logger.error("Error calling on startup function:\n{}".format(tb))
            raise SystemExit()
        # If it's a coroutine or otherwise awaitable, await it.
        # This is so that coroutines can be passed in to handle.
        if inspect.isawaitable(item):
            await item

    def on_startup(self, coro_or_callable: typing.Callable[['Kyoukai'], None]):
        """
        Registers a function to be called on startup.

        The function should be a callable.
        :return: The unmodified callable.
        """
        if not callable(coro_or_callable):
            raise TypeError("Object {} is not callable".format(coro_or_callable))
        self._on_startup = coro_or_callable

        return coro_or_callable

    async def handle_http_error(self, err: HTTPException, protocol, ctx: HTTPRequestContext):
        """
        Handle a :class:`kyoukai.exc.HTTPException`.

        This will invoke the appropriate error handler as registered in the blueprint of the route, if we can.
        Otherwise, it will invoke the default error handler.
        """
        code = err.code
        route = err.route
        # Check if the route is None.
        # If it is, we just have to use the default blueprint to handle the exception.
        if not route:
            bp = self._root_bp
        else:
            bp = route.bp

        # Get the error handler.
        error_handler = bp.get_errorhandler(code)
        if not error_handler:
            # Since there's no special error handler derived for this code, return a basic Response.
            # If it's a 500 and we're in debug mode, format the traceback.
            if err.code == 500 and self.debug:
                # Use the Kyoukai debugger.
                should_err, r = self._debugger.debug(ctx, err)
                protocol.handle_resp(r)
                return should_err
            else:
                body = str(code)
            resp = Response(code, body)
        else:
            # Invoke the error handler specified.
            resp = wrap_response(await error_handler.invoke(ctx, exception=err))
        protocol.handle_resp(resp)

        # Check if we should close the connection.
        if hasattr(ctx.request, "should_keep_alive"):
            if not ctx.request.should_keep_alive:
                protocol.close()
        else:
            # If it's that bad, just close it anyway.
            protocol.close()

        return True

    async def delegate_request(self, protocol, ctx: HTTPRequestContext):
        """
        Handles a :class:`kyoukai.context.HTTPRequestContext` and it's underlying request, processing it to the route
        handlers and such in the blueprints.

        This is an **internal** method, and should not be used outside of the protocol, or for testing.
        """
        async with ctx:
            # Acquire the lock on the protocol.
            async with protocol.lock:
                # Check if we should skip our own handling and go straight to the debugger.
                if self.debug:
                    if '__debugger__' in ctx.request.args and ctx.request.args["__debugger__"] == "yes":
                        resp = self._debugger.debug(ctx, None)
                        protocol.handle_resp(resp[1])
                        return
                # Check if there's a host header.
                if ctx.request.version == (1, 1):
                    host = ctx.request.headers.get("host", None)
                    if not host:
                        exc = HTTPException(400)
                        self.log_request(ctx, code=400)
                        await self.handle_http_error(exc, protocol, ctx)
                        return
                # First, try and match the route.
                try:
                    route = self._match_route(ctx.request.path, ctx.request.method)
                except HTTPException as e:
                    # We matched it; but the route doesn't work for this method.
                    # So we catch the 405 error,
                    if e.code == 405:
                        self.log_request(ctx, code=e.code)
                        await self.handle_http_error(e, protocol, ctx)
                        return
                    else:
                        self.logger.error("??????? Something went terribly wrong.")
                        return

                # If the route did not match, return a 404.
                if not route:
                    fof = HTTPException(404)
                    self.log_request(ctx, code=404)
                    await self.handle_http_error(fof, protocol, ctx)
                    return

                # Try and invoke the Route.
                try:
                    # Note that this will already be a Response.
                    # The route should call `app._wrap_response` when handling the response.
                    # This is because routes are responsible for pre-route and post-route hooks, calling them in the
                    # blueprint as appropriate.
                    # So we just pass ourselves to the route and hope it invokes properly.
                    response = await route.invoke(ctx)
                except HTTPException as e:
                    # Handle a HTTPException normally.
                    self.log_request(ctx, e.code)
                    # Set the route of the exception.
                    e.route = route
                    await self.handle_http_error(e, protocol, ctx)
                    return
                except Exception as e:
                    # An uncaught exception has propogated down to our level - oh dear.
                    # Catch it, turn it into a 500, and return.
                    exc = HTTPException(500)
                    # Set the cause of the HTTP exception. Useful for 500 error handlers.
                    exc.__cause__ = e
                    # Set the route of the exception.
                    exc.route = route
                    self.log_request(ctx, 500)
                    should_err = await self.handle_http_error(exc, protocol, ctx)
                    if should_err:
                        self.logger.exception("Unhandled exception in route `{}`:".format(repr(route)))
                    return
                else:
                    # If there is no error happening, just log it as normal.
                    self.log_request(ctx, response.code)

                # Respond with the response.
                protocol.handle_resp(response)

                # Check if we should Keep-Alive it.
                if not ctx.request.should_keep_alive:
                    protocol.close()

    async def start(self, ip="0.0.0.0", port=4444, component=None):  # pragma: no cover
        """
        Run the Kyoukai component asynchronously.

        This bypasses Asphalt's runner completely and starts Kyoukai as it's own context.

        :param ip: The IP address to bind to.
        :param port: The port to bind to.
        :param component: The component to set on the application.
            If this is not passed in, it will create an empty one.
        """
        self.logger.warning("Kyoukai is bypassing Asphalt - contexts will not work.")
        ctx = Context()
        if not component:
            from kyoukai.asphalt import KyoukaiComponent
            self.component = KyoukaiComponent(self, ip, port)
        await self.component.start(ctx)

    def run(self, ip="0.0.0.0", port=4444, component=None):  # pragma: no cover
        """
        Runs the Kyoukai server from within your code.

        This is not normally invoked - instead Asphalt should invoke the Kyoukai component.
        However, this is here for convenience.
        """
        if not component:
            from kyoukai.asphalt import KyoukaiComponent
            component = KyoukaiComponent(self, ip, port)
        run_application(component)



"""
Module for Kyokai routes.
"""
import re

from asphalt.core import Context

import kyoukai
from kyoukai.context import HTTPRequestContext
from kyoukai.util import wrap_response
from kyoukai.converters import _converters, convert_args


class Route(object):
    """
    A route is simply a wrapped coroutine object for a request.

    :param blueprint: The blueprint this route is associated with.
    :param matcher: The regular expression to match routes against.
    :param methods: A :class:`list` of methods that are allowed in this route.
    :param bound: Internal. Used for views.
    :param run_hooks: If the request should run pre and post hooks, or be exempt.
    """

    def __init__(self, blueprint,
                 matcher: str, methods: list,
                 bound: bool = False,
                 run_hooks: bool = True):
        """
        Create a new Route.
        """
        self._match_str = matcher
        self.matcher = re.compile(matcher)
        self.allowed_methods = methods
        self._wrapped_coro = None

        self.bp = blueprint

        # Used for bounded routes.
        self._bound = bound

        self._view_class = None

        self.name = "<Undefined route>"

        self._should_run_hooks = run_hooks

    @property
    def self(self):
        if self._bound:
            return self._view_class
        else:
            return

    @self.setter
    def self(self, vc):
        if self._bound:
            self._view_class = vc
        else:
            raise TypeError("Attempted to update view class on unbounded route")

    def kyokai_match(self, path: str):
        """
        Check if a given path matches the specified route.
        """
        # If it's a hard match, do an lower == match
        matched = self.matcher.fullmatch(path)
        return matched

    def kyokai_method_allowed(self, meth: str):
        """
        Check if the method matches.
        """
        if meth.lower() == "head":
            return True
        meths = [m.lower() for m in self.allowed_methods]
        if 'any' in meths:
            return True
        in_m = meth.lower() in meths
        return in_m

    def create(self, coro):
        """
        Set the internal coroutine state from the passed in value.
        """
        self._wrapped_coro = coro
        self.name = coro.__name__
        return self

    def __repr__(self):
        return "<Route '{}' in blueprint {}>".format(self.name, repr(self.bp))

    def __call__(self, coro):
        """
        Sets the coroutine.
        """
        return self.create(coro)

    async def invoke(self, ctx: HTTPRequestContext, exception=None):
        """
        Invoke the route, calling the underlying coroutine.
        """
        if self._should_run_hooks:
            # Run pre-request hooks.
            hooks = self.bp.get_pre_hooks(ctx)
            if hooks:
                for hook in hooks:
                    # Await the hook.
                    ctx = await hook(ctx)
                    if not isinstance(ctx, Context):
                        # idc about the subtype, as long as it's a context.
                        raise TypeError("Hook {} returned non-context".format(hook.__name__))

        m_obj = self.matcher.fullmatch(ctx.request.path)
        if m_obj:
            matches = m_obj.groups()
        else:
            matches = []

        # Invoke the coroutine.
        # Construct the params.
        params = []
        if self._bound:
            params.append(self.self)

        params.append(ctx)

        if exception:
            # This is an error handler, so add the exception.
            params.append(exception)

        if matches:
            params += list(matches)

        # Convert the arguments.
        args = convert_args(ctx, self._wrapped_coro, *params, bound=self._bound)

        result = await self._wrapped_coro(*args)
        # Wrap the result.
        result = wrap_response(result, response_cls=ctx.app.response_cls)

        if self._should_run_hooks:
            hooks = self.bp.get_post_hooks(ctx)
            if hooks:
                for hook in hooks:
                    result = await hook(ctx, result)
                    if not isinstance(result, kyoukai.Response):
                        raise TypeError("Hook {} returned non-response".format(hook.__name__))
                    result = wrap_response(result)

        # Set the request object.
        result.request = ctx.request

        return result

    def match(self, route: str, method: str):
        """
        Short way of checking if the route matches.

        This does *not* check if the method matches.
        """
        matched = self.kyokai_match(route)
        if matched:
            return self



"""
.. versionadded:: 1.8

The Kyoukai debugger is a subset of the Werkzeug debugger.

It uses the Werkzeug debugger for the heavy lifting, whilst providing this as a wrapper around to ensure that it acts
properly inside the very different Kyoukai HTTP server.
"""
import mimetypes
import os
import typing

from werkzeug.debug import get_current_traceback
from werkzeug import debug

from kyoukai.context import HTTPRequestContext
from kyoukai.response import Response


class KyoukaiDebugger:
    """
    The main debugger class.

    You should never create this yourself. Only the application object should create this.
    """

    def __init__(self, app):
        self.app = app

        self.frames = {}
        self.tracebacks = {}

    def debug(self, ctx: HTTPRequestContext, exc: Exception) -> typing.Tuple[bool, Response]:
        """
        Produces debug output for the application in a new template.
        """
        if not self.app.debug:
            return

        # Check the request's params.
        params = ctx.request.args

        debugger = params.get("__debugger__")
        if debugger == "yes":
            command = params.get("cmd")
            # Switch based on the command.
            if command is None:
                return Response(code=404, body="404")
            elif command == "resource":
                # Send a resource down the line.
                filename = params.get("f")
                # Get the __file__ of the werkzeug debugger.
                wz_f = os.path.dirname(debug.__file__)
                filename = os.path.join(wz_f, 'shared', os.path.basename(filename))
                # Guess the content type from the filename.
                mimetype = mimetypes.guess_type(filename)[0] or 'application/octet-stream'
                if os.path.exists(filename):
                    with open(filename, 'rb') as f:
                        return False, Response(body=f.read(), code=200, headers={"Content-Type": mimetype})
                else:
                    return False, Response(body="404", code=404)
            else:
                # It's a console command.
                frame = self.frames.get(int(ctx.request.args.get('frm')))
                if frame is None:
                    return False, Response(body="404", code=404)
                # Run the console command inside the frame.
                result = frame.console.eval(command)
                return False, Response(body=result, code=200, headers={"Content-Type": "text/html"})
        else:
            # Get the traceback, now.
            traceback = get_current_traceback(
                skip=0, show_hidden_frames=True,
                ignore_system_exceptions=False)
            for frame in traceback.frames:
                self.frames[frame.id] = frame
            self.tracebacks[traceback.id] = traceback
            # Render the base page.
            body = traceback.render_full(evalex=True, evalex_trusted=True, )
            r = Response(code=500, body=body.encode(), headers={"X-XSS-Protection": "0"})
            return True, r



"""
Class that handles logins.
"""
import types
import typing

import functools

import asyncio

from kyoukai import Kyokai, Request
from kyoukai.context import HTTPRequestContext
from kyoukai.response import redirect, Response

try:
    import itsdangerous
except ImportError:
    raise ImportError("itsdangerous is not installed - cannot use session extension")


class SessionUser(object):
    """
    This is a stub class, that is used for KyoukaiSession.

    You can override this class and use it with KyoukaiSession.
    """
    def __init__(self):
        """
        The only attribute that is required here is `self.id`.
        """
        self.id = None


class KyoukaiSession(object):
    """
    A KyoukaiSession object provides a few useful methods for handling client logins.

    The class takes a callable, which will then be called to get a user upon a request.
    The callable must take a HTTPRequestContext, or other suitable class.

    This user will be passed in to your route, as the second argument.

    To require login on a route, decorate the route with `@sess.login_required`, for example:

    ```python
    @app.route("/")
    @session.login_required
    async def index(ctx: HTTPRequestContext, user: User):
        return "You are " + repr(user)
    ```

    If the user is not logged in, then it will redirect them to `redirect_uri`. This can be passed as the second param.

    This ext is Flask-like in terms of handling the application object - you can either pass it as the last item of
    the args, or call `init_app` later to initialize it.

    To login a user, return `session.login(id, redirect_uri="/")`, which will return a response that logs in the
    specified ID, and redirects them.
    """
    def __init__(self, user_f, redirect_uri: str="/login", app: Kyokai=None):
        """
        Create a new KyoukaiSession object.
        """

        self._callable_f = user_f
        self.redirect_uri = redirect_uri

        self.secret_key = b""
        self.signer = None

        if app:
            self.init_app(app)

    def init_app(self, app: Kyokai):
        """
        Loads the secret key from the config.
        """
        try:
            self.secret_key = app.config["secret_key"].encode()
        except KeyError:
            raise Exception("You must set a secret key in `config.yml` for sessions to work.")

        self.signer = itsdangerous.Signer(self.secret_key)

    def _get_id(self, request: Request):
        """
        Checks a request for the cookie headers.
        """
        cook = request.cookies.get("KySess")
        if not cook:
            return None
        # Load, and unsign.
        try:
            id = self.signer.unsign(cook)
        except itsdangerous.BadSignature:
            return None
        else:
            return id

    def login_required(self, func):
        """
        Decorator for checking if the login is correct.
        """
        @functools.wraps(func)
        async def _login_required_fake_func(ctx: HTTPRequestContext, *args):
            """
            You will never see this.

            Enforces login.
            """
            id = self._get_id(ctx.request)
            if not id:
                return redirect(self.redirect_uri)
            # Get the user object.
            if asyncio.iscoroutine(self._callable_f):
                # Don't know why, but ok
                u = await self._callable_f
            elif asyncio.iscoroutinefunction(self._callable_f) or \
                    hasattr(self._callable_f, "__await__"):
                u = await self._callable_f(id)
            else:
                u = self._callable_f(id)
            # Await the underlying function.
            return await func(ctx, u, *args)

        return _login_required_fake_func

    def with_user(self, func):
        """
        Decorator for loading a user, but not requiring one.
        """
        @functools.wraps(func)
        async def _with_user_fake_func(ctx: HTTPRequestContext, *args):
            """
            You should never see this.
            """
            id = self._get_id(ctx.request)
            if id:
                if asyncio.iscoroutine(self._callable_f):
                    # Don't know why, but ok
                    u = await self._callable_f
                elif asyncio.iscoroutinefunction(self._callable_f) or \
                        hasattr(self._callable_f, "__await__"):
                    u = await self._callable_f(id)
                else:
                    u = self._callable_f(id)
            else:
                u = None
            # await the fnc
            return await func(ctx, u, *args)

        return _with_user_fake_func

    def login(self, id, redirect_uri="/"):
        """
        Logs in a user, and returns a Response.
        """
        r = redirect(redirect_uri)
        r.cookies["KySess"] = self.signer.sign(id)
        return r


"""
A request represents a client wanting to get a resource from the server.

This is automatically passed into your app route when an appropriate path is recieved.
"""
import json
import urllib.parse as uparse

from http import cookies

# Use the C parser if applicable.
from io import BytesIO

from werkzeug.datastructures import Accept, Authorization
from werkzeug.exceptions import ClientDisconnected

try:
    from http_parser.parser import HttpParser, IOrderedDict
except ImportError:
    from http_parser.pyparser import HttpParser, IOrderedDict

# Used for parsing headers.
from werkzeug import http, formparser

from kyoukai.exc import HTTPException


class Request(object):
    """
    A Request object.

    This should not be manually created. Instead, it is automatically provided by Kyokai.

    If you must create one, use :meth:`from_data` or :meth:`parse`.

    :ivar method: The HTTP method (GET, POST, PUT, etc)
    :ivar path: The full path of the request (``/api/v1/whatever``)
    :ivar headers: A :class:`IOrderedDict` representing the headers of the request.
    :ivar query: The raw query string (``a=b&c=d``)
    :ivar body: The raw body of the request.

    :ivar cookies: A :class:`cookies.SimpleCookie` containing the cookies of the request.
    :ivar raw_data: The raw data of the request.

    :ivar args: The arguments from the query string, parsed out.
    :ivar form: The form data for the request. If the request was JSON, this is automatically parsed out.
    :ivar values: THe arguments and the form combined.

    :ivar source: The source IP of the request.
    """

    def _parse(self, parser: HttpParser):
        """
        Parse the data.
        """
        self._parser = parser
        # Check if fully parsed before continuing.
        if not self.fully_parsed:
            # This is due to Kyoukai's buffer handling.
            # The request is called to parse multiple times.
            # Now, we just return if we're not 100% ready, meaning we won't get parsed needlessly.
            # This saves CPU and memory parsing stuff.
            return
        self.version = parser.get_version()
        self.method = parser.get_method()
        self.path = uparse.unquote(parser.get_path())
        self.headers = parser.get_headers()
        self.query = parser.get_query_string()
        self.body = parser.recv_body()

        self.cookies = cookies.SimpleCookie()
        self.cookies.load(self.headers.get("Cookie", "")) or {}

        self.raw_data = b""

        self.source = "0.0.0.0"

        self.extra = {}

        self.should_keep_alive = parser.should_keep_alive()

        self.version = parser.get_version()
        self.sversion = '.'.join(map(str, self.version))

        # urlparse out the items.
        _raw_args = uparse.parse_qs(self.query, keep_blank_values=True)
        # Reparse args
        self.args = IOrderedDict()
        for k, v in _raw_args.items():
            if len(v) == 1:
                self.args[k] = v[0]
            elif len(v) == 0:
                self.args[k] = None
            else:
                self.args[k] = v
        if self.headers.get("Content-Type") == "application/json":
            # Parse as JSON
            try:
                self.form = json.loads(self.body.decode())
            except json.JSONDecodeError:
                # Malformed JSON.
                raise HTTPException(400)

            # JSON bodies obviously don't have "native" upload support.
            # If you want to use a files key in the JSON, that's fine.
            self.files = {}

        else:
            # Parse the form data out.
            f_parser = formparser.FormDataParser()

            # Wrap the body in a BytesIO.
            body = BytesIO(self.body)

            # The headers can't be directly passed into Werkzeug.
            # Instead, we have to get a the custom content type, then pass in some fake WSGI options.
            mimetype, c_t_args = http.parse_options_header(self.headers.get("Content-Type"))

            if not mimetype:
                # Ok, no body.
                self.form = {}
                self.files = {}

            else:
                # Construct a fake WSGI environment.
                env = {"Content-Type": self.headers.get("Content-Type"),
                       "Content-Length": self.headers.get("Content-Length")}

                # Take the boundary out of the Content-Type, if applicable.
                boundary = c_t_args.get("boundary")
                if boundary is not None:
                    env["boundary"] = boundary

                # Get a good content length.
                content_length = self.headers.get("Content-Length")
                try:
                    content_length = int(content_length)
                except ValueError:
                    content_length = len(self.body)
                except TypeError:
                    # NoneType...
                    raise HTTPException(411)

                # Then, the form body itself is parsed.
                try:
                    data = f_parser.parse(body,
                                          mimetype,
                                          content_length,
                                          options=env
                                          )
                except ClientDisconnected:
                    # This means we were asked to parse before the request is fully made.
                    # This is a quirk of how Kyoukai handles buffers.
                    # We can just safely return here.
                    return

                # Extract the new data from the form parser.
                self.form = data[1]
                self.files = data[2]

        self.values = IOrderedDict(self.args)
        self.values.update(self.form if self.form else {})

    @property
    def accept(self) -> Accept:
        """
        Parses the Accept header
        :return: A new :class:`werkzeug.datastructures.Accept` object.
        """
        header = self.headers.get("Accept", "")
        return http.parse_accept_header(header)

    @property
    def auth(self) -> Authorization:
        """
        Parses the Authorization header.

        Note: this will return None in the case of an authorization header that is not Basic or Digest.
        :return: A new :class:`werkzeug.datastructures.Authorization` object.
        """
        header = self.headers.get("Authorization", "")
        return http.parse_authorization_header(header)

    @property
    def fully_parsed(self):
        """
        Checks if the request is fully parsed or not.

        If called inside user code, this is almost definitely True. Being not is probably a bug.

        :return: If the request has been fully parsed.
        """

        if not hasattr(self, "_parser"):
            return False
        return self._parser.is_message_complete()

    def parse(self, data: bytes, source: str):
        """
        Parse the request.

        :param data: The HTTP request's raw data.
        :param source: The source address of the
        """
        parser = HttpParser()
        data_len = len(data)
        # Execute the parser.
        parsed_len = parser.execute(data, data_len)
        if parsed_len == 0 or (data_len != parsed_len and parser.is_message_complete()):
            raise HTTPException(400, "Bad Request")

        self.raw_data = data
        self.source = source

        # Parse the data.
        self._parse(parser)

    @classmethod
    def from_data(cls, data: bytes, source: str):
        """
        Create a new request from request data.

        Shortcut for:

        .. code:: python

            r = Request()
            r.parse(data, source)
        """
        # Create a new request.
        req = cls()

        req.parse(data, source)

        return req



"""
Asphalt framework mixin for Kyokai.
"""
import logging

import asyncio
from functools import partial
from typing import Union

from asphalt.core import Component, resolve_reference, Context
from typeguard import check_argument_types

from kyoukai.app import Kyoukai
from kyoukai.protocol import KyoukaiProtocol
from kyoukai.context import HTTPRequestContext

logger = logging.getLogger("Kyoukai")


class KyoukaiComponent(Component):
    def __init__(self, app: Union[str, Kyoukai], ip: str = '0.0.0.0', port: int = 4444, **cfg):
        assert check_argument_types()
        if not isinstance(app, Kyoukai):
            self.app = resolve_reference(app)
        else:
            self.app = app
        self.ip = ip
        self.port = port
        self._extra_cfg = cfg

        # Reconfigure the app with the extra config.
        self.app.reconfigure(**self._extra_cfg)

        # Set HTTPRequestContext's `cfg` val to the extra config.
        HTTPRequestContext.cfg = self._extra_cfg

        self.server = None

    def get_protocol(self, ctx: Context):
        return KyoukaiProtocol(self.app, ctx)

    async def start(self, ctx: Context):
        """
        Starts a Kyokai server.
        """
        # Call on_startup.
        await self.app.call_on_startup()
        protocol = partial(self.get_protocol, ctx)
        self.server = await asyncio.get_event_loop().create_server(protocol, self.ip, self.port)
        logger.info("Kyoukai serving on {}:{}.".format(self.ip, self.port))



"""
Misc utilities.
"""
import os
import pathlib

from kyoukai.exc import HTTPException

VERSION = "1.8.0.1"
VERSIONT = tuple(map(int, VERSION.split('.')))


def static_filename(filename: str) -> str:
    """
    Naive static filename implementation, to allow serving static files.
    """
    built = ""
    p = pathlib.PurePath(filename)
    for part in p.parts:
        if part != "..":
            built += part + os.path.sep

    return built[:-1]


def wrap_response(response, response_cls=None):
    """
    Wrap up a response, if applicable.

    This allows Flask-like `return ""`.

    :param response: The tuple or otherwise object that is being wrapped.
    :param response_cls: The Response class that is being used.
    """
    # Import inside here to prevent circular imports.
    if response_cls is None:
        from kyoukai.response import Response
    else:
        Response = response_cls

    if response is None:
        r = Response(204, "", {})
    elif isinstance(response, tuple):
        if len(response) == 1:
            # Only body.
            r = Response(200, response[0], {})
        elif len(response) == 2:
            # Body and code.
            r = Response(response[1], response[0], {})
        elif len(response) == 3:
            # Body, code, headers.
            r = Response(response[1], response[0], response[2])
        else:
            # what
            raise HTTPException
    elif isinstance(response, Response):
        r = response
    else:

        r = Response(200, response, {})
    return r
